(*
 * Copyright (C) 2004 Damien Pous
 * 
 * This file is part of CCRecMin.
 * 
 * CCRecMin is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * CCRecMin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with CCRecMin; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *)

Require Import vector.
Require Export minski_tools.
Set Implicit Arguments.



Section Clear.

  (* Rk <- 0 *)

  Variables k n: nat.
  Hypothesis k_n: k < S n.

  Definition Clear: Minski 1 n := Test 1 n k k_n 1 0 (le_n _) (le_O_n _) :: nnil _. 

  Theorem ClearSemR: forall r r', r' = change r k 0 -> MinskiSemR [Clear] r r'.
  Proof.
    intros r r' Heq; rewrite Heq; generalize r; clear Heq r' r.
    cut (forall v r, v=get r k k_n -> MinskiSemR [Clear] r (change r k 0)).
    intros H r; apply (H _ r (refl_equal _)).
    intro v; induction v; intros r Hv;
    step_transition; rewrite <- Hv.
    transition_refl; rewrite (change_id k_n); auto.
    rewrite <- (change_ii_invol k 0 v); apply IHv; vauto.
  Qed.

End Clear.



Section Inkr.

  (* Rk++ *)

  Variables k n: nat.
  Hypothesis k_n: k < S n.

  Definition Inkr: Minski 1 n := Incr 1 n k k_n 1 (le_n _) :: nnil _. 

  Theorem InkrSemR: forall r r', r' = change r k (S (get r k k_n)) -> MinskiSemR [Inkr] r r'.
  Proof.
    intros r r' Heq; rewrite Heq.
    step_transition; apply transition_refl.
  Qed.

End Inkr.



Section Decr.

  (* Rk-- *)

  Variables k n: nat.
  Hypothesis k_n: k < S n.

  Definition Decr: Minski 1 n := Test 1 n k k_n 1 1 (le_n _) (le_n _) :: nnil _. 

  Theorem DecrSemR: forall r r', r' = change r k (pred (get r k k_n)) -> MinskiSemR [Decr] r r'.
  Proof.
    intros r r' Heq; rewrite Heq; generalize r; clear Heq r' r.
    cut (forall v r, v=get r k k_n -> MinskiSemR [Decr] r (change r k (pred v))).
    intros H r; apply (H _ r (refl_equal _)).
    intros v r Hv;
    step_transition; rewrite <- Hv.
    destruct v; simpl; [ rewrite (change_id k_n); auto | idtac ];
    transition_refl.
  Qed.

End Decr.



Section Move.

  Variables k k' n: nat.
  Hypothesis k_n:  k < S n.
  Hypothesis k'_n: k'< S n.


  (* Rk' <- Rk' + Rk    (et Rk <- 0) *)

  Definition MoveAdd: Minski 2 n := 
    Test 2 n k k_n 2 1 (le_n _) (le_n_Sn _) :: 
    Incr 2 n k' k'_n 0 (le_O_n _) :: 
    nnil _. 

  Theorem MoveAddSemR: k<>k' -> forall r r', r' = change (change r k 0) k' (get r k k_n + get r k' k'_n) -> 
    MinskiSemR [MoveAdd] r r'.
  Proof.
    intros k_k' r r' Heq; rewrite Heq; generalize r; clear Heq r' r.
    cut (forall v v' r, v=get r k k_n -> v'=get r k' k'_n -> MinskiSemR [MoveAdd] r (change (change r k 0) k' (v+v'))).
    intros H r; apply H; reflexivity. 
    intro v; induction v; intros v' r Hv Hv'; step_transition; rewrite <- Hv.
    transition_refl; simpl; rewrite (change_id k_n  _ Hv ); rewrite (change_id k'_n _ Hv'); reflexivity.
    step_transition;
    rewrite get_change_ij; try exact k_k';
    rewrite plus_n_Sm;
    rewrite <- (change_ii_invol k 0 v);
    rewrite <- (change_ii_invol k' (v + S v') (S v')).
    rewrite    (change_ij_comm k_k' 0).
    rewrite <- Hv';
    apply IHv;
    (try rewrite (get_change_ij (sym_not_eq k_k')); vauto).
  Qed.



  (* Rk' <- Rk    (et Rk <- 0) *)
  
  Definition Move: Minski 3 n := Concat (Clear k'_n) MoveAdd.

  Theorem MoveSemR: forall r r', r' = change (change r k' (get r k k_n)) k 0 -> 
    MinskiSemR [Move] r r'.
  Proof.
    intros r r' Heq; rewrite Heq; clear Heq r'.

    unfold Move. 
    MinskiSemR. 
    apply ClearSemR; auto.

    destruct (eq_nat_dec k k') as [k_k'|k_k'].
    unfold MoveAdd; subst.
    step_transition; vauto.
    apply transition_refl.

    apply MoveAddSemR; auto.
    rewrite (change_ij_comm k_k'); vauto.
    rewrite get_change_ij; auto.
    rewrite <- plus_n_O; reflexivity.
  Qed.
 
End Move.



Section Copy.

  (* Rk' <- Rk    (Rk inchang, Rt <- 0) *)

  Variables k k' t n: nat.
  Hypothesis kk' : k <> k'.
  Hypothesis kt  : k <> t.
  Hypothesis k't : k'<> t.
  Hypothesis kn  : k  < S n.
  Hypothesis k'n : k' < S n.
  Hypothesis tn  : t  < S n.

  Let dup_k_k'_t: Minski 3 n.
  refine (
    Test _ _ k  _ 3 1 (le_n _) _ ::
    Incr _ _ k' _ 2   _          ::
    Incr _ _ t  _ 0   (le_O_n _) :: 
    nnil _
  ); auto with arith. Defined.
 
  Let dup_sem: forall v v' V r, 
     v =get r k  kn  -> 
     v'=get r k' k'n -> 
     V =get r t  tn  ->
       MinskiSemR [dup_k_k'_t] r (change (change (change r k 0) k' (v+v')) t (v+V)).
  Proof.
    unfold dup_k_k'_t, MinskiSemR; msns.
    induction v; intros v' V r Hv Hv' HV; 
    step_transition; rewrite <- Hv.
    transition_refl; msns.
    rewrite (change_id kn  _ Hv );
    rewrite (change_id k'n _ Hv');
    rewrite (change_id tn  _ HV ); 
    reflexivity.

    step_transition; rewrite (get_change_ij kk');
    step_transition; rewrite (get_change_ij k't);
                     rewrite (get_change_ij kt).
    rewrite <- Hv'; rewrite <- HV; do 2 rewrite plus_n_Sm.
    rewrite <- (change_ii_invol k 0 v);
    rewrite <- (change_ii_invol k' (v  + S v') (S v'));
    rewrite    (change_ij_comm kk' 0).
    rewrite <- (change_ii_invol t (v  + S V) (S V));
    rewrite    (change_ij_comm k't (v + S v'));
    rewrite    (change_ij_comm kt).
    apply IHv;
    do 3 ((rewrite get_change_ii; reflexivity) || 
          (rewrite get_change_ij; [ idtac | exact (sym_not_eq kk') || 
                                            exact (sym_not_eq kt)  || 
                                            exact (sym_not_eq k't) ])).
  Qed.


  Definition Copy: Minski 7 n := 
     Concat (Clear tn) 
    (Concat (Clear k'n)
    (Concat  dup_k_k'_t
            (MoveAdd tn kn))).

  Theorem CopySemR: forall r r', r' = change (change r k' (get r k kn)) t 0 -> MinskiSemR [Copy] r r'.
  Proof.
    intros r r' Heq.
    unfold Copy; MinskiSemR.
    apply ClearSemR;   auto.
    apply ClearSemR;   auto.
    apply dup_sem;     auto.
    apply MoveAddSemR; auto.

    vauto.
    do 2 get_change.
    do 2 rewrite <- plus_n_O.
    rewrite (change_ij_comm kk'); vauto.
    rewrite (change_ij_comm kt ); vauto.
    rewrite (change_ij_comm (sym_not_eq k't)); vauto.
    rewrite <- (change_ij_comm kt );
    rewrite <- (change_ij_comm kk');
    rewrite (change_id kn); auto.
  Qed.

End Copy.



Section MultipleCopy.

  (* Rn+i <- Ri   pour i<m,    Rn+m <- 0 *)

  Let CopyOmSm m n: Minski 7 (S n + m).
  refine (
    fun m n => Copy (k:=0) (k':=m) (t:=S m) (n:=S n+m) (lt_O_Sn _) _ _
  ); simpl; auto with arith. 
  Defined.
    

  Fixpoint MultipleCopy (n m: nat) {struct n}: Minski (S (n*7)) (n+m) :=
    match n return Minski (S (n*7)) (n+m) with
    | 0 => Clear (lt_n_Sn _)
    | S n => Concat (CopyOmSm m n) (Shift (MultipleCopy n m) (le_n_Sn _))
    end.

  Definition mcopy n m (r r': vector nat (S n + m)): Prop :=
    (forall i, i<m -> forall h,    get r'  i    h = get r i h)  /\
    (forall i, i<n -> forall h h', get r' (i+m) h = get r i h') /\
    (                 forall h,    get r' (n+m) h = 0).

  Let in_i_Si_eq: forall i j, i<j -> j<=S i -> j=S i.
  Proof. intros i j H H'; apply le_antisym; auto with arith. Qed.
  Let lt_neq: forall i j, i<j -> j<>i.
  Proof. intros i j H F; rewrite F in H; elim (lt_irrefl _ H). Qed.

  Theorem MultipleCopySemR: forall n m (h: n<=m) r r', mcopy r r' ->
    MinskiSemR [MultipleCopy n m] r r'.
  Proof.
    induction n; intros m h r r' H;
    elim H; clear H; intros H1 H2; elim H2; clear H2; intros H2 H3;
    simpl.

    apply ClearSemR.
    apply vector_eq; intros i hi.
    destruct (eq_nat_dec m i) as [Hi|Hi].
    subst; vauto; rewrite <- H3; vauto.
    rewrite (get_change_ij Hi).
    apply H1; auto with arith.
    lt_eq_lt i m im; 
    [ assumption 
    | subst; elim (Hi (refl_equal _)) 
    | elim (lt_irrefl _ (le_lt_trans _ _ _ (lt_le_S _ _ im) hi)) ].

    cut (m<>0); intro mO; [ idtac | rewrite mO in h; elim (le_Sn_O _ h) ].
    cut (forall i, (S (S (n + m)) - S (n + m) + i) = S i); 
     [ intro B | rewrite <- minus_Sn_m; auto; rewrite <- minus_n_n; reflexivity ].

    MinskiSemR.

    unfold CopyOmSm; apply CopySemR; auto.

    apply IHn; auto with arith.
    split.
    intros i i_m i_Snm; vauto.
    do 2 (rewrite (get_pirr' (nki_n (le_n_S _ _ (le_n_Sn (n + m))) i_Snm) (lt_n_S _ _ i_Snm)); try apply B).
    rewrite get_change_ij.
    destruct (le_lt_dec m (S i)) as [m_Si|m_Si].
    generalize (in_i_Si_eq i_m m_Si); intro; subst; vauto; apply (H2 0 (lt_O_Sn _)).
    rewrite (H1 (S i) m_Si); rewrite (get_change_ij (lt_neq m_Si)); reflexivity.
    apply lt_neq; auto with arith.

    split.
    intros i i_n im_Snm i_Snm; vauto.
    destruct (le_lt_dec m i) as [m_i|m_i].
    elim (lt_irrefl _ (lt_trans _ _ _ i_n (lt_le_trans _ _ _ (lt_S_n _ _ (le_lt_n_Sm _ _ h)) m_i))).
    rewrite (get_pirr' (nki_n (le_n_S _ _ (le_n_Sn (n + m))) im_Snm) (lt_n_S _ _ im_Snm)); try apply B.
    rewrite (get_pirr' (nki_n (le_n_S _ _ (le_n_Sn (n + m))) i_Snm) (lt_n_S _ _ i_Snm)); try apply B.
    rewrite (get_change_ij (not_eq_S _ _ (lt_neq m_i))).
    destruct (eq_nat_dec m (S i)) as [m_Si|m_Si].
    subst; elim (lt_irrefl _ (le_lt_trans _ _ _ (le_S_n _ _ h) i_n)).
    rewrite (get_change_ij m_Si).
    apply (H2 (S i) (lt_n_S _ _ i_n)).

    intro; vauto.
    destruct m; [ elim (le_Sn_O _ h) | vauto ].
    rewrite (get_pirr' (nki_n (le_n_S _ _ (le_n_Sn (n + S m))) h0) (lt_n_Sn _)); try apply B.
    apply H3.

    rewrite <- minus_Sn_m; auto; rewrite <- minus_n_n.
    intros j hj hj' hj''; destruct j.
    do 2 (rewrite get_change_ij; auto).
    rewrite H1; auto.
    apply lt_le_trans with (S n); auto with arith.
    elim (lt_n_O _ (lt_S_n _ _ hj'')).

  Qed.

End MultipleCopy.


Ltac MiscSemR :=
  match goal with
  | |- MinskiSemR [Clear _] _ _          => apply ClearSemR
  | |- MinskiSemR [Inkr _] _ _           => apply InkrSemR
  | |- MinskiSemR [Decr _] _ _           => apply DecrSemR
  | |- MinskiSemR [Copy _ _ _] _ _       => apply CopySemR
  | |- MinskiSemR [MultipleCopy _ _] _ _ => apply MultipleCopySemR
  | |- MinskiSemR [Move _ _] _ _         => apply MoveSemR
  end.
