(*
 * Copyright (C) 2004 Damien Pous
 * 
 * This file is part of CCRecMin.
 * 
 * CCRecMin is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * CCRecMin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with CCRecMin; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *)

Require Import vector.
Require Export minski.
Set Implicit Arguments.



Section Nil.

  (* Machine vide *)

  Variables n: nat.

  Definition Nil: Minski 0 n := nnil _.

  Theorem NilSemR: forall r r': vector nat (S n), r=r' -> MinskiSemR [Nil] r r'.
  Proof. intros r r' H; rewrite <- H; unfold MinskiSemR; transition_refl. Qed.

End Nil.



Section Widen.

  (* largissemnt d'une machine  k registre en une machine  n registres *)

  Variables m k n: nat.
  Variable M: Minski m k.

  Hypothesis k_n: k<=n.
  Let Sk_Sn: S k<=S n. Proof le_n_S _ _ k_n.

  Let Widen_map(s: Step m k): Step m n :=
    match s with 
    | Incr i hi q    hq     => Incr _ _ i (lt_le_trans _ _ _ hi Sk_Sn) q    hq 
    | Test i hi q q' hq hq' => Test _ _ i (lt_le_trans _ _ _ hi Sk_Sn) q q' hq hq' 
    end.

  Definition Widen: Minski m n := map Widen_map M.

  Let unwiden(C: Config m n): Config m k := config m k (state C) (proof C) (truncate (S k) (reg C) Sk_Sn).
  Let widen(C: Config m k)(x0: vector nat (S n)): Config m n := config m n (state C) (proof C) (overwrite (reg C) x0).

  Let unwiden_widen: forall C x0, unwiden (widen C x0) = C. 
  Proof. intros C x0; unfold unwiden, widen, reg; rewrite (deConfig C); vauto. Qed.
  Hint Resolve unwiden_widen.

  Let Widen_step: forall C H C', 
    unwiden C' = step [M] (unwiden C) H -> eq_from (S k) (reg C) (reg C') -> 
    step [Widen] C H = C'.
  Proof.
    destruct C; destruct C'; unfold Widen, step, unwiden, state, reg, proof, Ms; 
    do 2 rewrite Minskiss'. 
    rewrite get_map.
    MsNs M m k; MsNs (map Widen_map M) m n.
    destruct (get M q H); unfold Widen_map; intros Hs Hr.
    apply deConfig; simpl.
    rewrite (state_eq Hs); reflexivity.
    apply (truncate_eq_from_eq Sk_Sn).
    rewrite (reg_eq Hs); vauto. 
    apply (eq_from_change Sk_Sn); assumption. 

    rewrite get_truncate in Hs.
    destruct (get r i (lt_le_trans i (S k) (S n) hi Sk_Sn));
    ( apply deConfig; simpl; 
      [ rewrite (state_eq Hs); reflexivity
      | apply (truncate_eq_from_eq Sk_Sn); 
        [ rewrite (reg_eq Hs); vauto | try assumption ]
      ] 
    ).
    apply (eq_from_change Sk_Sn); assumption. 
  Qed.

  Let WidenSem': forall (Cu Cu': Config m k), transition [M] Cu Cu' ->  
    forall C C', Cu=unwiden C -> Cu'=unwiden C' -> eq_from (S k) (reg C) (reg C') -> transition [Widen] C C'.
  Proof.
    intros Cu Cu' HCuCu'; induction HCuCu' as [Du|Du Du' Hnf H' HR]; intros C C' Hu Hu' Heq; subst.

    destruct C as [q h r];
    destruct C' as [q' h' r'].
    transition_refl;
    unfold unwiden in Hu'.
    exact (state_eq Hu').
    apply (truncate_eq_from_eq Sk_Sn).
    apply (reg_eq Hu').
    exact Heq.

    apply step_transition with Hnf. 
    rewrite (Widen_step Hnf (C':= widen (step [M] (unwiden C) Hnf) (reg C))); auto.
    apply HR; simpl; vauto.
    apply eq_from_overwrite; exact Heq.
    unfold widen; simpl; apply eq_from_sym; auto.
  Qed.
  
  Theorem WidenSemR: forall r r': vector nat (S n), 
    MinskiSemR [M] (truncate (S k) r Sk_Sn) (truncate (S k) r' Sk_Sn) -> 
    eq_from (S k) r r' -> 
    MinskiSemR [Widen] r r'.
  Proof.
    unfold MinskiSemR; intros r r' H1 H2.
    apply (WidenSem' H1).
    apply deConfig; auto.
    apply deConfig; auto.
    exact H2.
  Qed.

  Theorem WidenSem: forall k' (x: vector nat k') y, MinskiSem [M] x y -> MinskiSem [Widen] x y.
  Proof.
    intros k' x y H.
    apply SemR; intro x0.
    ElimMinskiSem (H (truncate (S k) x0 Sk_Sn)) r H1 H2.
    exists (overwrite r (overwrite x x0)); split. 
    simpl; apply WidenSemR.
    vauto.
    intros j hj hj' jSk; rewrite (get_overwrite_2 jSk); vauto.
    rewrite head_get; rewrite (get_overwrite_1 (lt_O_Sn (ns [M]))); reflexivity.
  Qed.

End Widen.



Section Shift.

  (* Dcalage d'une machine Machine  k registre en une machine  n registres *)

  Variables m k n: nat.
  Variable M: Minski m k.

  Hypothesis k_n: k<=n.
  Definition nk := n-k.

  Let Sk_Sn: S k<=S n. Proof le_n_S _ _ k_n.

  Let nki_Sn: forall i, i < S k -> nk+i < S n.
  Proof.
    intros; unfold nk.
    pattern n at 2; rewrite (le_plus_minus _ _ k_n).
    rewrite plus_comm; rewrite <- plus_Sn_m.
    apply plus_lt_compat_r; exact H.
  Qed.

  Let nkSk_Sn: S n = n-k+S k. 
  Proof. rewrite plus_comm; rewrite plus_Sn_m; rewrite <- (le_plus_minus _ _ k_n); reflexivity. Qed.

  Let Shift_map(s: Step m k): Step m n :=
    match s with 
    | Incr i hi q    hq     => Incr m n (n-k+i) (nki_Sn hi) q    hq 
    | Test i hi q q' hq hq' => Test m n (n-k+i) (nki_Sn hi) q q' hq hq' 
    end.

  Definition Shift: Minski m n := map Shift_map M.
  Let unshift(C: Config m n): Config m k := config m k (state C) (proof C) (etruncate (S k) (reg C) Sk_Sn).
  Let shift(C: Config m k) s: Config m n := config _ _ (state C) (proof C) (coerce _ nkSk_Sn (s @ (reg C))).

  Let unshift_shift: forall s C, unshift (shift C s) = C. 
  Proof. 
    intros s C; unfold unshift, shift; simpl; rewrite (deConfig C); vauto. 
    apply vector_eq; intros j h; vauto.
  Qed.
  Hint Resolve unshift_shift.

  Let Shift_step: forall C H C', 
    unshift C' = step [M] (unshift C) H -> eq_until (n-k) (reg C) (reg C') -> step [Shift] C H = C'.
  Proof.
    destruct C; destruct C'; unfold Shift, step, unshift, state, reg, proof, Ms; 
    do 2 rewrite Minskiss'. 
    rewrite get_map.
    MsNs M m k; MsNs (map Shift_map M) m n.
    destruct (get M q H); unfold Shift_map; intros Hs Hr.
    apply deConfig; simpl.
    rewrite (state_eq Hs); reflexivity.
    apply (etruncate_eq_until_eq Sk_Sn).
    rewrite (reg_eq Hs); vauto. 
    change (n-k) with (S n-S k); vauto.
    apply eq_until_change; auto with arith; apply le_minus.

    rewrite get_etruncate in Hs; change (S n-S k) with (n-k) in Hs;
    rewrite (le_pirr (nki_n Sk_Sn hi) (nki_Sn hi)) in Hs.
    destruct (get r (n - k + i) (nki_Sn hi));
    ( apply deConfig; simpl; 
      [ rewrite (state_eq Hs); reflexivity
      | apply (etruncate_eq_until_eq Sk_Sn); 
        [ rewrite (reg_eq Hs); change (n-k) with (S n-S k); vauto | try assumption ]
      ] 
    ).
    apply eq_until_change; auto with arith. apply le_minus. 
  Qed.

  Let ShiftSem': forall (Cu Cu': Config m k), transition [M] Cu Cu' ->
    forall C C', Cu = unshift C -> Cu' = unshift C' -> eq_until (n-k) (reg C) (reg C') -> transition [Shift] C C'.
  Proof.
    intros Cu Cu' HCuCu'; induction HCuCu' as [Du|Du Du' Hnf H' HR]; intros C C' Hu Hu' Heq; subst.

    destruct C as [q h r];
    destruct C' as [q' h' r'].
    transition_refl;
    unfold unshift in Hu'.
    exact (state_eq Hu').
    apply (etruncate_eq_until_eq Sk_Sn).
    apply (reg_eq Hu').
    exact Heq.

    apply step_transition with Hnf. 
    rewrite (Shift_step Hnf (C':= shift (step [M] (unshift C) Hnf) (truncate (n-k) (reg C) (le_S _ _ (le_minus _ _))))); auto.
    apply HR; simpl; vauto.
    intros j hj hj' nkj; vauto; rewrite (get_append_1 nkj); vauto.
    intros j hj hj' nkj; unfold shift, reg at 2; vauto; rewrite (get_append_1 nkj); vauto.
  Qed.

  Theorem ShiftSemR: forall r r': vector nat (S n), 
    MinskiSemR [M] (etruncate (S k) r Sk_Sn) (etruncate (S k) r' Sk_Sn) -> 
    eq_until (n-k) r r' -> 
    MinskiSemR [Shift] r r'.
  Proof.
    unfold MinskiSemR; intros r r' H1 H2.
    apply (ShiftSem' H1).
    apply deConfig; auto.
    apply deConfig; auto.
    exact H2.
  Qed.

End Shift.



Section Prepend.

  (* Ajout de code au dbut d'une machine *)

  Variables m m' n: nat.
  Variable M: vector (Step (m'+m) n) m'. 
  Variable N: Minski m n.

  Definition Prepend_map(s: Step m n): Step (m'+m) n :=
    match s with 
    | Incr i hi q    hq     => Incr (m'+m) n i hi (m'+q)         (plus_le_compat_l _ _ _ hq) 
    | Test i hi q q' hq hq' => Test (m'+m) n i hi (m'+q) (m'+q') (plus_le_compat_l _ _ _ hq) (plus_le_compat_l _ _ _ hq')  
    end.

  Definition Prepend: Minski (m'+m) n := M @ (map Prepend_map N).

  Lemma prepend_proof: forall i (hi: i<=m'+m), i-m'<=m.
  Proof.
    intros i hi.
    apply plus_le_reg_l with m'; auto.
    destruct (le_lt_dec m' i).
    rewrite <- (le_plus_minus m' i); auto with arith.
    rewrite not_le_minus_0; auto with arith.
  Qed.

  Let prepend(C: Config m n) := config (m'+m) n (m' + state C) (plus_le_compat_l _ _ _ (proof C)) (reg C).
  Definition unprepend(C: Config (m'+m) n) := config m n (state C - m') (prepend_proof (proof C)) (reg C).

  Let prepend_nfinal: forall C, nfinal C -> nfinal (prepend C).
  Proof. destruct C; unfold prepend, nfinal, state; intros; apply plus_lt_compat_l; assumption. Qed.
  Lemma unprepend_nfinal: forall C, nfinal (unprepend C) -> nfinal C.
  Proof. 
    destruct C; unfold unprepend, nfinal, state; intros. 
    apply le_lt_trans with (m'+(q-m')); auto with arith. 
    destruct (le_lt_dec m' q); auto with arith.
    rewrite le_plus_minus_r; auto with arith.
  Qed.

  Let prepend_step: forall C (hm: state C >= m') (H: nfinal (unprepend C)), 
    step [N] (unprepend C) H = unprepend (step [Prepend] C (unprepend_nfinal H)).
  Proof.
    intros C hm H.
    unfold step, Prepend, Ms; do 2 rewrite Minskiss'.
    MsNs N m n; MsNs (M @ map Prepend_map N) (m'+m) n.
    rewrite (get_append_2 hm).
    rewrite get_map.
    replace (get N (state C - m') (lt_minus m hm (unprepend_nfinal H))) with (get N (state (unprepend C)) H).
    replace (reg (unprepend C)) with (reg C).
    destruct (get N (state (unprepend C)) H); unfold Prepend_map;
    [ idtac | destruct (get (reg C) i hi) ]; unfold unprepend;
    apply deConfig; unfold state; auto with arith.
    reflexivity.
    apply get_pirr.
  Qed.

  Let prepend_step_m:  forall C (hm: state C >= m') (H: nfinal (unprepend C)), state (step [Prepend] C (unprepend_nfinal H)) - m' <= m.
  Proof.
    intros C hm H; unfold step, Prepend, Ms; rewrite Minskiss'.
    MsNs (M @ map Prepend_map N) (m'+m) n.
    rewrite (get_append_2 hm);
    rewrite get_map.
    destruct (get N (state C - m') (lt_minus m hm (unprepend_nfinal H))); unfold Prepend_map;
    [ idtac | destruct (get (reg C) i hi) ]; simpl; rewrite minus_plus; assumption.
  Qed.

  Let prepend_step_m': forall C (hm: state C >= m') (H: nfinal (unprepend C)), state (step [Prepend] C (unprepend_nfinal H)) >= m'.
  Proof.
    intros C hm H; unfold step, Prepend, Ms; rewrite Minskiss'.
    MsNs (M @ map Prepend_map N) (m'+m) n.
    rewrite (get_append_2 hm);
    rewrite get_map.
    destruct (get N (state C - m') (lt_minus m hm (unprepend_nfinal H))); unfold Prepend_map;
    [ idtac | destruct (get (reg C) i hi) ]; simpl; auto with arith.
  Qed.

  Let prepend_step_m'': forall (C C': Config (m'+m) n), transition [Prepend] C C' -> state C >= m' -> state C' >= m'.
  Proof.
    intros C C' H; induction H; intro B.
    exact B.
    apply IHtransition.
    unfold step, Prepend, Ms; rewrite Minskiss'.
    MsNs  (M @ map Prepend_map N) (m'+m) n. 
    rewrite (get_append_2 B); rewrite get_map.
    MsNsH Prepend (m'+m) n C.
    destruct (get N (state C - m') (lt_minus m B H)); simpl;
    [ idtac | destruct (get (reg C) i hi) ]; simpl; auto with arith.
  Qed.

  Theorem PrependSem: forall (Cu Cu': Config m n),
    forall (C: Config (m'+m) n) (hm: state C >= m') (C': Config (m'+m) n) (hm': state C' >= m'), 
    transition [N] Cu Cu' -> Cu = unprepend C -> Cu' = unprepend C' -> transition [Prepend] C C'.
  Proof.
    intros Cu Cu' C hm C' hm' H; gclear4 C hm C' hm';
    induction H as [Du|Du Du' Hnf H' HR]; intros C hm C' hm'.

    unfold unprepend;
    destruct C as [q hq r];
    destruct C' as [q' hq' r'];
    intros; subst.
    injection H0; clear H0; intros H1 H2; rewrite H1; 
    unfold state in hm; gclear hq; replace q with q'.
    intro hq; rewrite (le_pirr hq hq'); 
    apply transition_refl.
    rewrite (le_plus_minus _ _ hm); rewrite H2; rewrite <- le_plus_minus; auto. 

    intros; subst.
    apply step_transition with (unprepend_nfinal Hnf).
    apply HR; 
    [ apply prepend_step_m'; exact hm
    | exact hm'
    | apply prepend_step; exact hm
    | reflexivity ].
  Qed.

End Prepend.



Section Append.

  (* Ajout de code  la fin d'une machine *)

  Variables m m' n: nat.
  Variable M: Minski m n.
  Variable N: vector (Step (m+m') n) m'. 

  Definition Append_map(s: Step m n): Step (m+m') n :=
    match s with 
    | Incr i hi q    hq     => Incr (m+m') n i hi q    (le_plus_trans _ _ _ hq) 
    | Test i hi q q' hq hq' => Test (m+m') n i hi q q' (le_plus_trans _ _ _ hq) (le_plus_trans _ _ _ hq') 
    end.

  Definition Append: Minski (m+m') n := (map Append_map M) @ N. 

  Let appent(C: Config m n) := config (m+m') n (state C) (le_plus_trans _ _ _ (proof C)) (reg C). 
  Definition unappend(C: Config (m+m') n) (hq: state C <= m) := config m n (state C) hq (reg C).

  Let h_append: forall C, state (appent C) <= m.
  Proof. destruct C; auto with arith. Qed.
  Let unappend_append: forall C, unappend (appent C) (h_append C) = C.
  Proof. intro C; unfold unappend, append; rewrite (deConfig C); reflexivity. Qed.

  Let append_nfinal: forall C, nfinal C -> nfinal (appent C).
  Proof. destruct C; unfold append, nfinal, state; intros. apply lt_plus_trans; assumption. Qed.
  Let unappend_nfinal: forall C h, nfinal (unappend C h) -> nfinal C.
  Proof. destruct C; unfold unappend, nfinal; intros; apply lt_plus_trans; assumption. Qed.

  Let append_step: forall C h (H: nfinal (unappend C h)) P, step [M] (unappend C h) H = unappend (step [Append] C (append_nfinal H)) P.
  Proof.
    intros C h H.
    cut (nfinal C). intro H'.
    unfold step, Append, Ms; do 2 rewrite Minskiss';
    MsNs M m n; MsNs (map Append_map M @ N) (m+m') n.
    rewrite (get_append_1 H).
    rewrite get_map.
    replace (get M (state (unappend C h)) H) with (get M (state C) H).
    replace (reg (unappend C h)) with (reg C).
    destruct (get M (state C) H); unfold Append_map;
    [ idtac | destruct (get (reg C) i hi) ]; unfold unappend; intro P; apply deConfig; reflexivity.
    reflexivity.
    apply get_pirr.
    apply unappend_nfinal with h; exact H.
  Qed.

  Let append_step_m: forall C h (H: nfinal (unappend C h)), state (step [Append] C (append_nfinal H)) <= m.
  Proof.
    intros C h H; unfold step, Append, Ms; rewrite Minskiss';
    MsNs (map Append_map M @ N) (m+m') n.
    rewrite (get_append_1 H);
    rewrite get_map.
    destruct (get M (state (unappend C h)) H); unfold Append_map.
    assumption.
    destruct (get (reg C) i hi); assumption.
  Qed.

  Theorem AppendSem: forall (Cu Cu': Config m n),
    forall (C: Config (m+m') n) h (C': Config (m+m') n) h', 
    transition [M] Cu Cu' -> Cu = unappend C h -> Cu' = unappend C' h' -> transition [Append] C C'.
  Proof.
    intros Cu Cu' C h C' h' H; gclear4 C h C' h';
    induction H as [Du|Du Du' Hnf H' HR]; intros C h C' h'.

    unfold unappend;
    destruct C as [q hq r];
    destruct C' as [q' hq' r'];
    intros; subst.
    injection H0; intros; subst; rewrite (le_pirr hq hq'); apply transition_refl.

    intros; subst.
    apply step_transition with (append_nfinal Hnf).
    apply HR with (append_step_m Hnf) h'; [ apply append_step | reflexivity ].
  Qed.

End Append.



Section Concat.

  (* Concatnation de deux machines *)

  Variables m1 m2 n: nat.
  Variable M1: Minski m1 n.
  Variable M2: Minski m2 n.

  Definition Concat: Minski (m1+m2) n := Prepend (map (Append_map (m:=m1) (n:=n) m2) M1) M2.

  Theorem ConcatSemR: forall r' r r'': vector nat (S n), 
    MinskiSemR [M1] r r' -> MinskiSemR [M2] r' r'' -> MinskiSemR[Concat] r r''.
  Proof.
    unfold MinskiSemR, Concat; msns.
    intros r' r r'' rr' r'r''.
    apply transition_transition with (config (m1 + m2) n m1 (le_plus_l _ _) r').

    unfold Prepend; fold (Append M1 (map (Prepend_map (m:=m2) m1 (n:=n)) M2)).
    eapply AppendSem; auto.
    unfold unappend; apply rr'.

    eapply PrependSem; auto.
    simpl; unfold ge; apply le_plus_l.
    unfold unprepend. simpl. 
    replace (config m2 n (m1 - m1) (prepend_proof m2 m1 (le_plus_l m1 m2)) r') with (config m2 n 0 (le_O_n m2) r').
    replace (config m2 n (m1 + m2 - m1) (prepend_proof m2 m1 (le_n (m1 + m2))) r'') with (config m2 n m2 (le_n m2) r'').
    apply r'r''.
    apply deConfig; simpl; auto with arith.
    apply deConfig; simpl; auto with arith.
  Qed.

End Concat.

Ltac MinskiSemR :=
  match goal with
  | |- MinskiSemR [Nil] _ _            => apply NilSemR
  | |- MinskiSemR [Shift ?M ?h] _ _    => apply (ShiftSemR (M:=M) (k_n:=h)); try MinskiSemR
  | |- MinskiSemR [Widen ?M ?h] _ _    => apply (WidenSemR (M:=M) h); try MinskiSemR
  | |- MinskiSemR [Concat ?M1 ?M2] _ _ => eapply (ConcatSemR (M1:=M1) (M2:=M2)); try MinskiSemR
  end.



Section Loop.

  (* Mise en boucle de la machine M, tant que Rk>0 (avec dcrmentation automatique de Rk) *)

  Variables m n k: nat.
  Variable M: Minski m n.
  Hypothesis k_n: k<S n. 

  Definition Loop: Minski (1+(m+1)) n :=
    Prepend (Test _ _ k k_n (1+(m+1)) 1 (le_n _) (le_n_S _ _ (le_O_n _)) :: nnil _)
    (Append  M 
            (Test _ _ k k_n (m+1) 0 (le_n _) (le_O_n _) :: nnil _ )).

  Theorem LoopSemO: forall r: vector nat (S n), get r k k_n = 0 -> forall r', r=r' -> MinskiSemR [Loop] r r'.
  Proof.
    intros r Hrk r' Hrr'; unfold MinskiSemR; msns.
    unfold Loop, Prepend.
    step_transition; rewrite Hrk.
    transition_refl.
  Qed.

  Let mO_m: m-0 <= m. Proof. rewrite <- minus_n_O; apply le_n. Qed.

  Theorem LoopSemS: forall (r: vector nat (S n)) x, get r k k_n = S x -> 
    forall r' r'': vector nat (S n), 
    MinskiSemR [M] (change r k x) r' -> MinskiSemR [Loop] r' r'' -> MinskiSemR [Loop] r r''.
  Proof.
    intros r x Hrk r' r'' H1 H2.
    unfold MinskiSemR; msns.
    apply transition_transition with (config (S (m + 1)) n (S m) (le_plus_l _ _) r').

    step_transition.
    rewrite Hrk.
    unfold Loop.
    change (S (m+1)) with (1+(m+1)).
    eapply PrependSem; unfold state; auto with arith.
    unfold unprepend; simpl.
    eapply (fun N => AppendSem (m':=1) (M:=M) N (Cu:=config m n 0 (le_O_n _) (change r k x)) (Cu':=config m n m (le_n _) r')).
    apply H1.
    unfold unappend; simpl; reflexivity.
    unfold unappend; simpl.
    apply (deConfig (config m n (m-0) mO_m r')); unfold state; auto with arith.

    nfinal (config (S (m + 1)) n (S m) (le_plus_l _ _) r').
    apply step_transition with H.
    step_simpl.
    rewrite get_map.
    unfold Append.
    rewrite (get_append_2 (le_n m)).
    rewrite (get_pirr' (lt_minus 1 (le_n m) (lt_S_n m (m + 1) H)) (lt_O_Sn _)); simpl.
    inversion H2; clear C C' H4 H5.
    unfold Loop at 2, Prepend, step, Ms, state in H3; rewrite Minskiss' in H3.
    simpl in H3; msnsH H3.
    destruct (get r' k k_n); 
    [ rewrite (le_pirr (plus_le_compat_l (m + 1) (m + 1) 1 (le_n (m + 1))) (le_n (S (m + 1))))
    | rewrite (le_pirr (plus_le_compat_l 0 (m + 1) 1 (le_O_n (m + 1))) (le_n_S 0 (m + 1) (le_O_n (m + 1))))
    ]; exact H3.

    rewrite <- minus_n_n; reflexivity.
    apply lt_n_S; rewrite plus_comm; auto with arith. 
  Qed. 

End Loop.
Implicit Arguments Loop [m n].
