(*
 * Copyright (C) 2004 Damien Pous
 * 
 * This file is part of CCRecMin.
 * 
 * CCRecMin is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * CCRecMin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with CCRecMin; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *)

Require Export Arith.

Set Implicit Arguments.

(* Indiscernabilit des preuves de le, lt *)
Section le_pirr.
  Axiom le_pirr: forall i j (h h': i<=j), h=h'. 
  Theorem lt_pirr: forall i j (h h': i<j), h=h'. 
  Proof. unfold lt; intros; apply le_pirr. Qed.
  Hint Immediate lt_pirr le_pirr.
End le_pirr.

Ltac lt_eq_lt x y H := destruct (lt_eq_lt_dec x y) as [H|H]; [ destruct H as [H|H] | idtac ].

Ltac gclear  x := generalize x; clear x.
Ltac gclear2 x y := generalize x y; clear x y.
Ltac gclear3 x y z := generalize x y z; clear x y z.
Ltac gclear4 x y z t := generalize x y z t; clear x y z t.


(* tactique pour la rduction des termes sur les vecteurs *)
Ltac vauto := autorewrite with vector; auto with arith.

Section Vector.

  Variable A: Set.

  Inductive vector: nat -> Set :=
    | nnil: vector 0
    | ncons: forall (n: nat) (a: A), vector n -> vector (S n).
  
  Definition head (n: nat) (l: vector (S n)): A :=
    match l in vector k return 0<k -> A with 
    | nnil => fun h => False_rec A (lt_n_O _ h)
    | ncons _ a _ => fun _ => a 
    end (lt_O_Sn n).
  
  Definition tail (n: nat) (l: vector (S n)): vector n :=
    match l in vector k return 0<k -> vector (pred k) with 
    | nnil => fun h => False_rec (vector (pred 0)) (lt_n_O _ h)
    | ncons _ _ q => fun _ => q
    end (lt_O_Sn n).
  
  Fixpoint get (n: nat) (l: vector n) (k: nat) {struct l}: k<n -> A :=
    match l in vector n return k < n -> A with
    | nnil => fun h => False_rec _ (lt_n_O _ h)
    | ncons n a q => 
    	match k return k < S n -> A with 
    	| 0 => fun _ => a
  	| S k => fun h => get q (lt_S_n _ _ h)
  	end
    end.
  Implicit Arguments get [n].

  Fixpoint append (n: nat) (l: vector n) {struct l}: forall m: nat, vector m -> vector (n+m) :=
    match l in vector n return forall m: nat, vector m -> vector (n+m) with
    | nnil => fun _ l' => l'
    | ncons _ a q => fun _ l => ncons a (append q l)
    end.

  Fixpoint split1 (n: nat): forall (k: nat) (l: vector (n+k)), vector n :=
    match n return forall k (l: vector (n+k)), vector n with 
    | 0 => fun k l => nnil
    | S n => fun k l => ncons (head l) (split1 n k (tail l))
    end.
 
  Fixpoint split2 (n: nat): forall (k: nat) (l: vector (n+k)), vector k :=
    match n return forall k (l: vector (n+k)), vector k with 
    | 0 => fun k l => l
    | S n => fun k l => split2 n k (tail l)
    end.
 
  Fixpoint overwrite (k n: nat) (l: vector k) (l': vector n) {struct l}: vector n :=
    match l, l' with
    | nnil, l' => l'
    | ncons _ _ _, nnil => nnil
    | ncons _ a q, ncons _ a' q' => ncons a (overwrite q q')
    end.

  Fixpoint truncate (k: nat): forall n, vector n -> k<=n -> vector k :=
    match k return forall n, vector n -> k<=n -> vector k with
    | 0 => fun n l h => nnil
    | S k => fun n l => 
      match l in vector n return S k<=n -> vector (S k) with 
      | nnil => fun h => False_rec _ (le_Sn_O _ h)
      | ncons _ a q => fun h => ncons a (truncate q (le_S_n _ _ h))
      end
    end.
 
  Fixpoint rev_init (n: nat) (f: nat -> A) {struct n}: vector n :=
    match n return vector n with
    | 0   => nnil
    | S m => ncons (f n) (rev_init m f)
    end.

  Definition init (n: nat) (f: nat -> A) := rev_init n (fun m => f (n-m)).

(*
  Definition binit (n: nat) (f: forall i, i<n -> A): vector n.
    refine (
      fun n' f' => 
        (fix binit(n: nat) (f: forall i, i<n -> A) (k: nat) {struct k}: k<=n -> vector k :=
          match k return k<=n -> vector k with
          | 0   => fun _ => nnil
          | S k => fun h: S k<= n => ncons (f (n-S k) _) (binit n f k (lt_le_weak _ _ h))
          end
        ) n' f' n' (le_n _)
    ); auto with arith.
  Defined.
*)

  Definition fill(k n: nat) (l: vector k) (a: A) := 
    init n (fun i => match le_lt_dec k i with 
                     | right h => get l i h
                     | _ => a
                     end).

  Fixpoint change (n: nat) (l: vector n) (i: nat) (a: A) {struct l}: vector n :=
    match l in vector n, i return vector n with
    | nnil, _ => nnil
    | ncons _ a' q, 0 => ncons a q
    | ncons _ a' q, S i => ncons a' (change q i a)	
    end.	

  Lemma empty_nnil: forall l: vector 0, l=nnil.
  Proof.
    refine ((vector_ind (fun n => match n return vector n -> Prop with 0 => fun l => l=nnil | _ => fun _ => True end) (refl_equal nnil) _) 0).
    auto.  
  Qed.
    
  Lemma nempty_ncons: forall n (l: vector (S n)), exists x, exists q, l=ncons x q.
  Proof.
    intro n.
    refine ((vector_ind (fun n => match n return vector n -> Prop with S n => fun l => exists x, exists q, l=ncons x q | _ => fun _ => True end) _ _) (S n)).
    trivial.  
    intros m m' l H; exists m'; exists l; reflexivity.
  Qed.
    
  Theorem head_tail: forall (n: nat) (l: vector (S n)), l = ncons (head l) (tail l).
  Proof.
    intros n l.
    generalize (nempty_ncons l); intro H1; elim H1; intros x H2; elim H2; intros q H; clear H1 H2.
    rewrite H; simpl; reflexivity.
  Qed.

  
  Theorem vector_eq: forall (n: nat) (l l': vector n), 
  	(forall (i: nat) (h: i<n), get l i h = get l' i h) -> l=l'.
  Proof.
    induction l as [ | n x q HR ].
    intros; rewrite empty_nnil; reflexivity.
    intro l'; rewrite (head_tail l'); intro H.
    replace x with (head l').
    replace q with (tail l').
    reflexivity.
    rewrite (HR (tail l')).
    reflexivity.
    intros i h.
    generalize (H (S i) (lt_n_S _ _ h)).
    simpl; rewrite (lt_pirr (lt_S_n i n (lt_n_S i n h)) h); tauto.
    generalize (H 0 (lt_O_Sn _)).
    simpl; auto.
  Qed.

  Lemma get_cons: forall (n: nat) (a: A) (l: vector n) (i : nat) (h: i<n), 
                   get (ncons a l) (S i) (lt_n_S _ _ h) = get l i h.
  Proof.
    destruct l as [ | n a' q ]; intros i h.
    elim (lt_n_O _ h).
    simpl; rewrite (lt_pirr (lt_S_n i (S n) (lt_n_S i (S n) h)) h); reflexivity.
  Qed.

  Lemma head_get: forall (n: nat) (l: vector (S n)), head l = get l 0 (lt_O_Sn n).
  Proof. intros n l; rewrite (head_tail l); reflexivity. Qed.

  Definition eq_from(k n: nat) (l l': vector n): Prop := forall i (h h': i<n), k<=i -> get l i h = get l' i h'.
  Definition eq_until(k n: nat) (l l': vector n): Prop := forall i (h h': i<n), i<k -> get l i h = get l' i h'.
  Definition eq_between(k k' n: nat) (l l': vector n): Prop := forall i (h h': i<n), k<=i -> i<k' -> get l i h = get l' i h'.

  Theorem eq_from_sym: forall k n (l l': vector n), eq_from k l l' -> eq_from k l' l. 
  Proof. intros k n l l' H i h h' ki; rewrite (H i h' h); auto. Qed.

  Theorem eq_until_sym: forall k n (l l': vector n), eq_until k l l' -> eq_until k l' l. 
  Proof. intros k n l l' H i h h' ki; rewrite (H i h' h); auto. Qed.

  Theorem eq_between_sym: forall k k' n (l l': vector n), eq_between k k' l l' -> eq_between k k' l' l. 
  Proof. intros k k' n l l' H i h h' ki k'i; rewrite (H i h' h); auto. Qed.

End Vector.

Implicit Arguments get [A n].
Implicit Arguments truncate [A n].

Lemma get_pirr: forall i n (h h': i<n) A (l: vector A n), get l i h = get l i h'.
Proof.
  intros i n h h' A l; gclear3 i h h';
  induction l as [ | n x q HR].
  intros i h h'.
  elim (lt_n_O _ h).
  destruct i; simpl; intros; [ reflexivity | apply (HR i) ].
Defined.
Hint Immediate get_pirr.

Lemma get_pirr': forall i j n (h: i<n) (h': j<n) A (l: vector A n), i=j -> get l i h = get l j h'.
Proof. intros; subst; apply get_pirr. Qed.

Theorem get_tail: forall A (n: nat) (l: vector A (S n)) (i : nat) (h: i<n), get (tail l) i h = get l (S i) (lt_n_S _ _ h).
Proof. intros n A l i h; rewrite (head_tail l); simpl; apply get_pirr. Qed.

Theorem get_append_1: forall i n (h: i<n) A (l: vector A n) m (l': vector A m) (h': i<n+m), 
	        get (append l l') i h' = get l i h.
Proof. 
  intros i n h A l; gclear2 i h; 
  induction l as [ | n a q HR]; simpl; intros.
  elim (lt_n_O _ h).
  destruct i; [ reflexivity | rewrite <- (HR i (lt_S_n _ _ h) m l' (lt_S_n _ _ h')); apply get_pirr ].
Qed.

Lemma lt_minus: forall i j k (ij: i<=j) (jik: j<i+k), j-i<k.
Proof. intros i j k ij jik; apply plus_lt_reg_l with i; rewrite <- le_plus_minus; assumption. Qed.

Theorem get_append_2: forall i n (h: i>=n) A (l: vector A n) m (l': vector A m) (h': i<n+m), 
              get (append l l') i h' = get l' (i-n) (lt_minus m h h').
Proof. 
  intros i n h A l; gclear2 i h; 
  induction l as [ | n a q HR]; simpl; intros.
  apply get_pirr'; apply minus_n_O.
  destruct i.
  elim (le_Sn_O _ h).
  rewrite (get_pirr (lt_minus m h h') (lt_minus m (le_S_n _ _ h) (lt_S_n _ _ h')));
  simpl; apply HR. 
Qed.

Theorem get_append_2': forall i m n (h: m+i<m+n) A (l: vector A m) (l': vector A n), 
              get (append l l') (m+i) h = get l' i (plus_lt_reg_l _ _ _ h).
Proof. 
  intros i m n h A l l';
  rewrite (get_append_2 (le_plus_l m i));
  apply get_pirr';
  auto with arith.
Qed.
Hint Rewrite get_append_2': vector.

Theorem get_split1: forall n k i (h: i<n) A (l: vector A (n+k)), get (split1 n k l) i h = get l i (lt_plus_trans _ _ _ h).
Proof.
  induction n; intros k i h A l; simpl.
  elim (lt_n_O _ h).
  destruct i.
  rewrite (head_tail l); reflexivity.
  rewrite IHn; rewrite (head_tail l); simpl; apply get_pirr'; reflexivity.
Qed.
Hint Rewrite get_split1: vector.

Theorem get_split2: forall n k i (h: i<k) A (l: vector A (n+k)), get (split2 n k l) i h = get l (n+i) (plus_lt_compat_l _ _ _ h).
Proof.
  induction n; intros k i h A l; simpl; auto.
  rewrite IHn; simpl; rewrite (head_tail l); simpl; auto.
Qed.
Hint Rewrite get_split2: vector.

Theorem split: forall A n k (l: vector A (n+k)), l = append (split1 n k l) (split2 n k l).
Proof.
  intros; apply vector_eq; intros i h.
  destruct (le_lt_dec n i) as [ni|ni].
  rewrite (get_append_2 ni); vauto; apply get_pirr'; auto with arith.
  rewrite (get_append_1 ni); vauto.
Qed.
Hint Immediate split.

Theorem split1_append: forall m n A (l: vector A m) (l': vector A n), split1 m n (append l l') = l.
Proof. intros; apply vector_eq; intros i hi; vauto; apply (get_append_1 hi). Qed.
Hint Rewrite split1_append: vector.

Theorem split2_append: forall m n A (l: vector A m) (l': vector A n), split2 m n (append l l') = l'.
Proof. intros; apply vector_eq; intros i hi; vauto; rewrite (get_append_2 (le_plus_l m i)); apply get_pirr'; auto with arith. Qed.
Hint Rewrite split2_append: vector.

Theorem split_eq: forall m n A (l l': vector A (m+n)), split1 m n l = split1 m n l' -> split2 m n l = split2 m n l' -> l = l'.
Proof.
  intros m n A l l' H1 H2;
  rewrite (split m n l);
  rewrite (split m n l');
  rewrite H1;
  rewrite H2;
  reflexivity.
Qed.
Ltac split_eq m n := match goal with |- ?l = ?l' => rewrite (split m n l); rewrite (split m n l'); apply split_eq end. 

Theorem get_overwrite_1: forall i k (h': i<k) n (h: i<n) A (l : vector A k) (l': vector A n), get (overwrite l l') i h = get l i h'.
Proof.
  induction i; intros k h' n h A l l';
  (destruct l; [ elim (lt_n_O _ h') | destruct l'; [ elim (lt_n_O _ h) | simpl; auto ]]).
Qed.

Theorem get_overwrite_2: forall i k (h': k<=i) n (h: i<n) A (l : vector A k) (l': vector A n), get (overwrite l l') i h = get l' i h.
Proof.
  induction i; intros k h' n h A l l'; destruct l; destruct l'; simpl; auto.
  elim (le_Sn_O _ h').
  apply IHi.
  exact (le_S_n _ _ h').
Qed.

Theorem split1_overwrite: forall m n A (l: vector A m) (l': vector A (m+n)), split1 m n (overwrite l l') = l.
Proof.
  intros; apply vector_eq; intros i hi; vauto.
  apply get_overwrite_1. 
Qed.
Hint Rewrite split1_overwrite: vector.

Theorem split1_overwrite': forall k m n A (l: vector A k) (l': vector A (m+n)), split1 m n (overwrite l l') = (overwrite l (split1 _ _ l')).
Proof.
  intros; apply vector_eq; intros i hi; vauto.
  destruct (le_lt_dec k i) as [ki|ki].
  do 2 (rewrite get_overwrite_2; vauto).
  do 2 rewrite (get_overwrite_1 ki); vauto. 
Qed.
Hint Rewrite split1_overwrite: vector.

Theorem split2_overwrite: forall m n A (l: vector A m) (l': vector A (m+n)), split2 m n (overwrite l l') = split2 m n l'.
Proof.
  intros; apply vector_eq; intros i hi; vauto.
  apply get_overwrite_2; auto with arith.
Qed.
Hint Rewrite split2_overwrite: vector.

Theorem split2_overwrite': forall k m n A (l: vector A (m+k)) (l': vector A (m+n)), split2 m n (overwrite l l') = (overwrite (split2 m k l) (split2 m n l')).
Proof.
  intros; apply vector_eq; intros i hi; vauto.
  destruct (le_lt_dec (m+k) (m+i)) as [ki|ki].
  do 2 (rewrite get_overwrite_2; vauto).
  apply plus_le_reg_l with m; exact ki.
  rewrite (get_overwrite_1 ki); vauto. 
  rewrite (get_overwrite_1 (plus_lt_reg_l _ _ _ ki)); vauto. 
Qed.
Hint Rewrite split2_overwrite': vector.

Theorem overwrite_nothing: forall n A (l: vector A 0) (l': vector A n), overwrite l l' = l'.
Proof. intros n A l l'; rewrite (empty_nnil l); reflexivity. Qed.
Hint Rewrite overwrite_nothing: vector.

Theorem overwrite_full: forall n A (l l': vector A n), overwrite l l' = l.
Proof. intros; apply vector_eq; intros; apply get_overwrite_1. Qed.
Hint Rewrite overwrite_full: vector.

Theorem overwrite_empty: forall n A (l: vector A n) (l': vector A 0), overwrite l l' = nnil _.
Proof. induction l; intro l'; try rewrite (empty_nnil l'); vauto. Qed.
Hint Rewrite overwrite_empty: vector.

Theorem get_truncate: forall A k i n (l: vector A n) (h: k<=n) (h': i<k), get (truncate k l h) i h' = get l i (lt_le_trans _ _ _ h' h).
Proof.
  induction k; intros; simpl.
  elim (lt_n_O _ h').
  destruct l.
  elim (lt_n_O _ (lt_le_trans _ _ _ h' h)).
  destruct i; simpl.
  reflexivity.
  rewrite IHk; vauto.
Qed.
Hint Rewrite get_truncate: vector.

Theorem truncate_id: forall A n (l: vector A n) (h: n<=n), truncate n l h = l.
Proof. induction l; intros; simpl; try rewrite IHl; vauto. Qed.
Hint Rewrite truncate_id: vector.

Theorem truncate_overwrite_id: forall A k n (l': vector A k) (l: vector A n) (h: k<=n), truncate k (overwrite l' l) h = l'.
Proof. 
  intros; apply vector_eq; intros j hj; vauto.
  rewrite (get_overwrite_1 hj); reflexivity.
Qed.
Hint Rewrite truncate_overwrite_id: vector.

Theorem change_append_1: forall A m n (l: vector A m) (l': vector A n) i a, i<m -> change (append l l') i a = append (change l i a) l'.
Proof. 
  induction l; intros. 
  elim (lt_n_O _ H).
  simpl; destruct i; [ reflexivity | rewrite IHl; auto with arith ].
Qed.

Theorem change_append_2: forall A m n (l: vector A m) (l': vector A n) i a, change (append l l') (m+i) a = append l (change l' i a).
Proof. induction l; intros; [ reflexivity | simpl; rewrite IHl; reflexivity ]. Qed.
Hint Rewrite change_append_2: vector.

Theorem get_change_ii: forall i n (h: i<n) A (l: vector A n) a,
                         get (change l i a) i h = a.
Proof.
  intros i n h A l; gclear2 i h;
  induction l; intros i h a'.
  elim (lt_n_O _ h).
  destruct i; [ reflexivity | simpl; apply IHl ].
Qed.
Hint Rewrite get_change_ii: vector.

Theorem get_change_ij: forall i j (d: i<>j) A n (h: j<n) (l: vector A n) a,
                         get (change l i a) j h = get l j h.
Proof.
  intros i j d A n h l; gclear4 i j d h;
  induction l.
  intros i j ij h a; elim (lt_n_O _ h).
  intros i j neq_ij h a'.
  destruct i; destruct j; try reflexivity.
  elim neq_ij; reflexivity.
  simpl; apply IHl.
  intro eq_ij; elim neq_ij; rewrite eq_ij; reflexivity.
Qed.

Theorem change_ii_invol: forall A i a a' n (l: vector A n),
                         change (change l i a') i a = change l i a.
Proof.
  intros A i a a' n l; apply vector_eq.
  intros j h; destruct (eq_nat_dec i j) as [H|H].
  rewrite H; vauto. 
  do 3 (rewrite get_change_ij; try exact H ); reflexivity.
Qed.
Hint Rewrite change_ii_invol: vector.

Theorem change_ij_comm: forall i j (d: i<>j) A ai aj n (l: vector A n),
                         change (change l i ai) j aj = change (change l j aj) i ai.
Proof.
  intros i j neq_i_j A a a' n l; apply vector_eq.
  intros k h; destruct (eq_nat_dec i k) as [H|H].
  rewrite H; rewrite get_change_ii; rewrite get_change_ij; 
  [ vauto | rewrite H in neq_i_j; exact (sym_not_eq neq_i_j) ].
  rewrite (get_change_ij H); destruct (eq_nat_dec j k) as [H'|H'].
  rewrite H'; vauto.
  do 3 (rewrite get_change_ij; try assumption); reflexivity.
Qed.

Theorem change_id: forall i n (h: i<n) A (l: vector A n) a, a=get l i h -> change l i a = l.
Proof.
  intros i n h A l a H; apply vector_eq.
  intros k hk; destruct (eq_nat_dec i k) as [H'|H'];
  [ rewrite H'; rewrite get_change_ii; subst; apply get_pirr 
  | rewrite get_change_ij; [ reflexivity | simpl in H'; apply H' ] ].
Qed.

Theorem change_id': forall i j n (hj: j<n) A (l: vector A n) a, i=j -> a=get l j hj -> change l i a = l.
Proof.
  intros i j n hj A l a Hij Heq; subst.
  rewrite (change_id hj); reflexivity.
Qed.

Theorem change_id_out: forall i n A (l: vector A n) a, change l (n+i) a = l.
Proof.
  intros; apply vector_eq; intros.
  rewrite get_change_ij.
  reflexivity.
  intro F; rewrite <- F in h.
  elim (lt_irrefl _ (le_lt_trans _ _ _ (le_plus_l _ _) h)). 
Qed.
Hint Rewrite change_id_out: vector.

Theorem head_change_O: forall A n (l: vector A (S n)) a, head (change l 0 a) = a.
Proof. intros A n l a; rewrite (head_tail l); reflexivity. Qed.
Hint Rewrite head_change_O: vector.

Theorem head_change_S: forall A n (l: vector A (S n)) i a, head (change l (S i) a) = head l.
Proof. intros A n l i a; rewrite (head_tail l); reflexivity. Qed.
Hint Rewrite head_change_S: vector.

Theorem split1_change: forall m n A (l: vector A (m+n)) i a, split1 m n (change l i a) = change (split1 m n l) i a.
Proof.
  intros; apply vector_eq; intros j hj; vauto.
  destruct (eq_nat_dec i j) as [ij|ij].
  subst; vauto.
  do 2 rewrite (get_change_ij ij); vauto.
Qed.
Hint Rewrite split1_change: vector.

Theorem split1_change': forall m n A (l: vector A (m+n)) i a, split1 m n (change l (m+i) a) = split1 m n l.
Proof. intros; apply vector_eq; intros; vauto. Qed.
Hint Rewrite split1_change': vector.

Theorem split2_change: forall m n A (l: vector A (m+n)) i a, split2 m n (change l (m+i) a) = change (split2 m n l) i a.
Proof.
  intros; apply vector_eq; intros j hj; vauto.
  destruct (eq_nat_dec i j) as [ij|ij].
  subst; vauto.
  rewrite (get_change_ij ij).
  cut (m+i<>m+j). intro mij; rewrite (get_change_ij mij); vauto.
  intro F; generalize (plus_reg_l _ _ _ F); assumption.
Qed.
Hint Rewrite split2_change: vector.

Theorem split2_change': forall m n A (l: vector A (m+n)) i a, i<m -> split2 m n (change l i a) = split2 m n l.
Proof.
  intros; apply vector_eq; intros j hj; vauto.
  destruct (eq_nat_dec i (m+j)) as [ij|ij].
  subst; elim (lt_irrefl _ (lt_plus_trans _ _ _ H)).
  apply (get_change_ij ij).
Qed.

Lemma get_rev_init: forall i n (h: i<n) (A: Set) (f: nat -> A),
                     get (rev_init n f) i h = f (n-i).
Proof.
  intros i n; gclear i.
  induction n as [ | n HR ]; intros i h A f.
  elim (lt_n_O _ h).
  destruct i.
  reflexivity.
  simpl; apply HR.
Qed.
Hint Rewrite get_rev_init: vector.

Theorem get_init: forall i n (h: i<n) (A: Set) (f: nat -> A),
                     get (init n f) i h = f i.
Proof.
  unfold init; intros; rewrite get_rev_init.
  replace (n-(n-i)) with i.
  reflexivity.
  destruct n.
  elim (lt_n_O _ h).
  induction i.
  simpl; rewrite <- minus_n_n; reflexivity.
  change (S n-S i) with (n-i).
  rewrite <- minus_Sn_m; rewrite <- (minus_Sn_m n i) in IHi; try exact (lt_le_weak _ _ (lt_S_n _ _ h)).
  pattern i at 1; rewrite IHi; auto with arith.
  apply le_minus.
Qed.
Hint Rewrite get_init: vector.

Theorem get_fill_1: forall i k (h': i<k) A n (h: i<n) (l: vector A k) a, get (fill n l a) i h = get l i h'.
Proof.
  intros i k h' A n h l a;
  unfold fill; rewrite get_init;
  destruct (le_lt_dec k i) as [H|H].
  elim (lt_irrefl _ (lt_le_trans _ _ _ h' H)).
  apply get_pirr.
Qed.

Theorem get_fill_2: forall A i k n (h: i<n) (l: vector A k) a, k<=i -> get (fill n l a) i h = a.
Proof.
  intros A i k n h l a h'.
  unfold fill; rewrite get_init;
  destruct (le_lt_dec k i) as [H|H].
  reflexivity.
  elim (lt_irrefl _ (lt_le_trans _ _ _ H h')).
Qed.

Theorem change_fill: forall A i k (h: i<k) n (l: vector A k) a a', change (fill n l a) i a' = fill n (change l i a') a.
Proof.
  intros A i k ik n l a a'.
  apply vector_eq; intros j h.
  destruct (le_lt_dec k j) as [kj|kj]; 
  [ rewrite get_fill_2; try exact kj 
  | rewrite (get_fill_1 kj) ];
  destruct (eq_nat_dec i j) as [ij|ij]; subst.
  elim (lt_irrefl _ (lt_le_trans _ _ _ ik kj)).
  rewrite (get_change_ij ij); rewrite get_fill_2; [ reflexivity | exact kj ].
  vauto.
  do 2 rewrite (get_change_ij ij); apply get_fill_1.
Qed.

Theorem append_fill_O: forall A (l l': vector A 0) m n a, append (fill m l a) (fill n l' a) = fill (m+n) (nnil _) a.
Proof.
  intros A l l' m n a.
  split_eq m n; vauto;
  ( apply vector_eq; intros i h; vauto; do 2 (rewrite get_fill_2; vauto) ).
Qed.
Hint Rewrite append_fill_O: vector.

Theorem split1_fill: forall A k (l: vector A k) m n a, split1 m n (fill (m+n) l a) = fill m l a.
Proof.
  intros A k l m n a.
  apply vector_eq; intros i h; vauto.
  destruct (le_lt_dec k i) as [ki|ki].
  rewrite (get_fill_2 h l a ki); apply get_fill_2; auto.
  do 2 rewrite (get_fill_1 ki); reflexivity.
Qed.
Hint Rewrite split1_fill: vector.

Theorem split2_fill: forall A k (l: vector A k) m n a, k<=m -> split2 m n (fill (m+n) l a) = fill n (nnil _) a.
Proof.
  intros A k l m n a H.
  apply vector_eq; intros i h; vauto.
  rewrite get_fill_2; vauto.
  rewrite get_fill_2; auto with arith.
Qed.

Theorem fill_overwrite: forall m n (h: m<=n) k A (l: vector A m) (l': vector A n) a, fill k (overwrite l l') a = overwrite l (fill k l' a).
Proof.
  intros m n h k A l l' a; apply vector_eq; intros i hi.
  destruct (le_lt_dec n i) as [ni|ni].
  rewrite get_fill_2; auto.
  rewrite get_overwrite_2; auto.
  rewrite get_fill_2; auto.
  apply le_trans with n; assumption.
  rewrite (get_fill_1 ni).
  destruct (le_lt_dec m i) as [mi|mi].
  do 2 rewrite (get_overwrite_2 mi); rewrite (get_fill_1 ni); reflexivity.
  do 2 rewrite (get_overwrite_1 mi); reflexivity.
Qed.

Theorem overwrite_invol: forall A m (l: vector A m) n (l': vector A n), overwrite l (overwrite l l') = overwrite l l'.
Proof.
  induction l; vauto.
  destruct l'; simpl; vauto.
  rewrite IHl; reflexivity.
Qed.
Hint Rewrite overwrite_invol: vector.

Theorem change_overwrite_1: forall i k (h: i<k) A n (l: vector A k) (l': vector A n) a, change (overwrite l l') i a = overwrite (change l i a) l'.
Proof.
  intros; apply vector_eq; intros.
  destruct (eq_nat_dec i i0).
  subst; rewrite (get_overwrite_1 h); vauto.
  rewrite (get_change_ij n0).
  destruct (le_lt_dec k i0).
  do 2 rewrite (get_overwrite_2 l0); reflexivity.
  do 2 rewrite (get_overwrite_1 l0); rewrite get_change_ij; vauto.
Qed. 

Theorem overwrite_append_1: forall m k (h: k<=m) A n (l: vector A k) (l1: vector A m) (l2: vector A n), overwrite l (append l1 l2) = append (overwrite l l1) l2.
Proof.
  intros; apply vector_eq; intros.
  destruct (le_lt_dec k i) as [ki|ki]; 
  destruct (le_lt_dec m i) as [mi|mi]; 
  repeat (rewrite (get_overwrite_2 ki) || rewrite (get_overwrite_1 ki) 
       || rewrite (get_append_2 mi) || rewrite (get_append_1 mi));
  vauto.
  elim (lt_irrefl _ (lt_le_trans _ _ _ ki (le_trans _ _ _ h mi))).
Qed. 
Hint Rewrite overwrite_append_1: vector.

Theorem change_overwrite_2: forall i k (h: k<=i) A n (l: vector A k) (l': vector A n) a, change (overwrite l l') i a = overwrite l (change l' i a).
Proof.
  intros; apply vector_eq; intros.
  destruct (eq_nat_dec i i0).
  subst; rewrite (get_overwrite_2 h); vauto.
  rewrite (get_change_ij n0).
  destruct (le_lt_dec k i0).
  do 2 rewrite (get_overwrite_2 l0); rewrite get_change_ij; vauto.
  do 2 rewrite (get_overwrite_1 l0); reflexivity.
Qed. 

Theorem truncate_append: forall A m n (l: vector A m) (l': vector A n) (h: m<=m+n), truncate m (append l l') h = l.
Proof.
  intros; apply vector_eq; intros.
  rewrite get_truncate; apply get_append_1.
Qed.
Hint Rewrite truncate_append: vector.

Theorem truncate_change: forall A n (l: vector A n) i k a (h: k<=n), truncate k (change l i a) h = change (truncate k l h) i a.
Proof.
  induction l; intros.
  generalize (le_n_O_eq k h); intro; subst; vauto.
  destruct i; destruct k; simpl; vauto.
  rewrite IHl; reflexivity.
Qed.
Hint Rewrite truncate_change: vector.

Theorem truncate_overwrite: forall A m (l: vector A m) n (l': vector A n) k (h:k<=n), truncate k (overwrite l l') h = overwrite l (truncate k l' h).
Proof.
  induction l; intros m l' k h; vauto.
  destruct l'.
  generalize (le_n_O_eq _ h); intro; subst; reflexivity.
  destruct k; simpl; try rewrite IHl; reflexivity.
Qed.
Hint Rewrite truncate_overwrite: vector.

Theorem truncate_eq_from_eq: forall k m (h: k<=m) A (l l': vector A m), truncate k l h = truncate k l' h -> eq_from k l l' -> l = l'.
Proof.
  intros k m h A l l' H1 H2; apply vector_eq; intros j hj.
  destruct (le_lt_dec k j) as [kj|kj].
  apply (H2 j); exact kj.
  cut (get (truncate k l h) j kj = get (truncate k l' h) j kj). 
  vauto; do 2 rewrite (get_pirr hj (lt_le_trans j k m kj h)); tauto.
  rewrite H1; reflexivity.
Qed.

Theorem eq_from_change: forall k m (h: k<=m) A i (l l': vector A m) a, i<k -> eq_from k l l' -> eq_from k (change l i a) l'.
Proof.
  intros k m h A i l l' a ik H j hj hj' kj.
  rewrite get_change_ij.
  apply H; exact kj.
  intro F; subst; elim (lt_irrefl _ (lt_le_trans _ _ _ ik kj)).
Qed.

Theorem eq_from_overwrite': forall A m n (l: vector A m) (l': vector A n), eq_from m (overwrite l l') l'.
Proof.
  intros A m n l l' j hj hj' mj.  
  rewrite get_overwrite_2; vauto.
Qed.
Hint Immediate eq_from_overwrite'.

Theorem eq_from_overwrite: forall A m n (l: vector A m) (l1 l2: vector A n), eq_from m l1 l2 -> eq_from m (overwrite l l1) l2.
Proof.
  intros A m n l l1 l2 H j hj hj' mj.  
  rewrite get_overwrite_2; vauto.
Qed.

Theorem eq_from_append: forall A m n (l1 l2: vector A m) (l: vector A n), eq_from m (append l1 l) (append l2 l).
Proof.
  intros A m n l1 l2 l j hj hj' mj; 
  do 2 rewrite (get_append_2 mj); vauto.
Qed.
Hint Immediate eq_from_append.
(*
Theorem eq_from_append: forall k m (h: m<=k) n A (l l': vector A m) (l1 l2: vector A n), eq_from (k-m) l1 l2 -> eq_from k (append l l1) (append l' l2).
Proof.
  intros k m mk n A l l' l1 l2 H j hj hj' mj.  
  do 2 rewrite (get_append_2 (le_trans _ _ _ mk mj)).
  apply H.
  apply plus_le_reg_l with m.
  do 2 (rewrite <- le_plus_minus; auto).
  apply le_trans with k; assumption.
Qed.
*)
Lemma nki_n: forall i k n, k<=n -> i < k -> n-k+i < n.
Proof.
  intros i k n kn ik.
  pattern n at 2; rewrite (le_plus_minus _ _ kn).
  rewrite plus_comm.
  apply plus_lt_compat_r; exact ik.
Qed.

Definition etruncate k n (l: vector nat n) (h: k<=n): vector nat k :=
  init k (fun i => match le_lt_dec k i with left _ => 0 | right h' => get l (n-k+i) (nki_n h h') end).
Implicit Arguments etruncate [n].

Theorem get_etruncate: forall k n (l: vector nat n) (h: k<=n) i (h': i<k), get (etruncate k l h) i h' = get l (n-k+i) (nki_n h h').
Proof.
  intros; unfold etruncate; vauto.
  destruct (le_lt_dec k i) as [ki|ki].
  elim (lt_irrefl _ (le_lt_trans _ _ _ ki h')).
  vauto.
Qed.
Hint Rewrite get_etruncate: vector.

Theorem etruncate_change: forall n (l: vector nat n) i k a (h: k<=n), etruncate k (change l ((n-k)+i) a) h = change (etruncate k l h) i a.
Proof.
  intros; apply vector_eq; intros j hj; vauto.
  destruct (eq_nat_dec i j) as [ij |ij].
  subst; vauto.
  rewrite (get_change_ij ij).
  rewrite get_change_ij; vauto.
  intro F; generalize (plus_reg_l _ _ (n-k) F); exact ij.
Qed.
Hint Rewrite etruncate_change: vector.

Theorem etruncate_append: forall m n (l: vector nat m) (l': vector nat n) (h: n<=m+n), etruncate n (append l l') h = l'.
Proof.
  intros; apply vector_eq; intros;
  rewrite get_etruncate. 
  rewrite (get_pirr' (nki_n h h0) (plus_lt_compat_l _ _ _ h0)); vauto.
  rewrite (plus_comm m n); auto with arith.
Qed.
Hint Rewrite etruncate_append: vector.

Theorem etruncate_append_cons: forall m n (l: vector nat m) (l': vector nat n) a (h: n<=m+S n), etruncate n (append l (ncons a l')) h = l'.
Proof.
  intros; apply vector_eq; intros.
  rewrite get_etruncate. 
  cut (m<=m + S n - n + i). 
  intro mm; rewrite (get_append_2 mm).
  rewrite (get_pirr' (lt_minus (S n) mm (nki_n h h0)) (lt_n_S _ _ h0)).
  destruct i; simpl; vauto.
  rewrite (plus_comm m); do 2 (rewrite plus_Sn_m; rewrite plus_n_Sm; rewrite minus_plus); reflexivity.
  rewrite (plus_comm m); rewrite plus_Sn_m; rewrite plus_n_Sm; rewrite minus_plus; auto with arith.
Qed.
Hint Rewrite etruncate_append_cons: vector.

Theorem etruncate_eq_until_eq: forall k m (h: k<=m) (l l': vector nat m), etruncate k l h = etruncate k l' h -> eq_until (m-k) l l' -> l = l'.
Proof.
  intros k m h l l' H1 H2; apply vector_eq; intros j hj.
  destruct (le_lt_dec (m-k) j) as [mkj|mkj].
  cut (j-(m-k)<k). intro hjmk.
  cut (get (etruncate k l h) (j-(m-k)) hjmk = get (etruncate k l' h) (j-(m-k)) hjmk). 
  vauto. 
  replace (get l  j hj) with (get l  (m - k + (j - (m - k))) (nki_n h hjmk)).
  replace (get l' j hj) with (get l' (m - k + (j - (m - k))) (nki_n h hjmk)).
  tauto.
  apply get_pirr'; rewrite <- (le_plus_minus _ _ mkj); reflexivity.
  apply get_pirr'; rewrite <- (le_plus_minus _ _ mkj); reflexivity.
  rewrite H1; reflexivity.
  apply lt_le_trans with (m-(m-k)).
  apply plus_lt_reg_l with (m-k).
  rewrite <- (le_plus_minus _ _ mkj); rewrite <- (le_plus_minus _ _ (le_minus m k)); exact hj.
  pattern m at 1; rewrite (le_plus_minus _ _ h); rewrite plus_comm; rewrite minus_plus; apply le_n.
  apply (H2 j); exact mkj.
Qed.

Theorem eq_until_change: forall k m (h: k<=m) A i (l l': vector A m) a, k<=i -> eq_until k l l' -> eq_until k (change l i a) l'.
Proof.
  intros k m h A i l l' a ik H j hj hj' kj.
  rewrite get_change_ij.
  apply H; exact kj.
  intro F; subst; elim (lt_irrefl _ (le_lt_trans _ _ _ ik kj)).
Qed.

Lemma eq_le: forall m n, m=n -> m<=n. Proof. intros; subst; apply le_n. Qed.

Definition coerce m n (h: n=m) (l: vector nat m) := etruncate n l (eq_le h).
Implicit Arguments coerce [m].

Theorem get_coerce: forall m n (h: n=m) (l: vector nat m) i (h': i<n), get (coerce n h l) i h' = get l i (lt_le_trans _ _ _ h' (eq_le h)).
Proof.
  unfold coerce; intros; vauto.
  apply get_pirr'; subst; rewrite <- minus_n_n; reflexivity.
Qed.
Hint Rewrite get_coerce: vector.


Section Map.

  Variables A B: Set.
  Variable f: A -> B.

  Fixpoint map(n: nat) (l: vector A n) {struct l}: vector B n :=
  match l in vector _ n return vector B n with
  | nnil => nnil B
  | ncons _ a q => ncons (f a) (map q)
  end.

  Theorem get_map: forall (n: nat) (l: vector A n) (i: nat) (h: i<n), get (map l) i h = f (get l i h).
  Proof.
    induction l.
    intros i h; elim (lt_n_O _ h).
    destruct i; [ reflexivity | intro h; simpl; apply IHl ].
  Qed.
  Hint Rewrite get_map: vector.

End Map.

(* notations *)
Infix "::" := ncons (at level 60, right associativity).
Infix "@" := append (at level 61, left associativity).

(* dcomposition automatique *)
Ltac finite_list l := rewrite (empty_nnil l) || ( rewrite (head_tail l); finite_list (tail l) ).
Ltac get_change := rewrite get_change_ii || (rewrite get_change_ij; [ get_change | assumption || auto with arith ]).
